/* 
 * $Id: ssl_client.c,v 1.11 2006-04-08 15:14:12+09 naito Exp root $
 * OpenSSL-based CAS ticket validator by Shawn Bayern, based on 
 * OpenSSL demonstration code.
 */

#include <stdio.h>
#include <memory.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include "cas.h"

#define END(x) { ret = (x); goto end; }
#define FAIL END(0)
#define FAIL_BAD_CERT END(BAD_CERT)
#define SUCCEED END(1)

static void log(const char *) ;

int CAS_validate_SSL(char *, char *, char *, int, mod_cas2_conf *) ;
static X509 *get_cert_from_file(char *);
static int valid_cert(X509 *, char *, char **, char **);

/** Returns status of certification:  0 for invalid, 1 for valid. */
static int valid_cert(X509 *cert, char *hostname, char **trusted_ca, char **trusted_ca_dir)
{
  X509_STORE *store ;
  X509_STORE_CTX *ctx ;
  X509_LOOKUP *lookup ;
  int err, i ;
  char buf[4096];

  ctx = X509_STORE_CTX_new();
  store = X509_STORE_new();
  lookup=X509_STORE_add_lookup(store,X509_LOOKUP_file());
  if (lookup == NULL) {
    LOG("lookup NULL\n") ;
    return -1 ;
  }
  for (i = 0; trusted_ca[i] != NULL; i++) {
    if (!X509_LOOKUP_load_file(lookup,trusted_ca[i],X509_FILETYPE_PEM)) {
      LOG(trusted_ca[i]) ; LOG("\n") ;
      X509_LOOKUP_load_file(lookup,NULL,X509_FILETYPE_DEFAULT);
    }
  }
  
  lookup=X509_STORE_add_lookup(store,X509_LOOKUP_hash_dir());
  if (lookup == NULL) { 
    LOG("lookup NULL\n") ;
    return -1 ; 
  }
  for (i = 0; trusted_ca_dir[i] != NULL; i++) {
    X509_LOOKUP_add_dir(lookup,trusted_ca_dir[i],X509_FILETYPE_PEM) ;
    LOG(trusted_ca_dir[i]) ; LOG("\n") ;
  }
  X509_LOOKUP_add_dir(lookup,NULL,X509_FILETYPE_DEFAULT);
  
  ERR_clear_error();

  if(!X509_STORE_CTX_init(ctx, store, NULL, NULL)) {
    LOG("ERROR\n") ; 
  }
  if (X509_verify_cert(ctx) == 0) {
    err = X509_STORE_CTX_get_error(ctx) ;
    switch (ctx->error) {
    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
      LOG("X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:\n") ; 
      break ;
    case X509_V_ERR_CERT_NOT_YET_VALID:
      LOG("X509_V_ERR_CERT_NOT_YET_VALID:\n") ; 
      break;
    case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
      LOG("case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:\n") ; 
      break;
    case X509_V_ERR_CERT_HAS_EXPIRED:
      LOG("case X509_V_ERR_CERT_HAS_EXPIRED:\n") ; 
      break;
    case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
      LOG("X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:\n") ;
      break;
    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
      LOG("X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:\n") ; 
      break;
    default:
      LOG("OTHER\n") ; 
      break;
    }
    return 0 ;
  }
  X509_NAME_get_text_by_NID(X509_get_subject_name(cert), NID_commonName, buf, sizeof(buf) - 1);
  if (strlen(hostname) >= sizeof(buf) - 1) return 0;
  return (!strcmp(buf, hostname));
}

/** Returns status of ticket by filling 'buf' with a NetID if the ticket
 *  is valid and buf is large enough and returning 1.  If not, 0 is
 *  returned.
 */
int CAS_validate_SSL(char *ticket, char *service, char *outbuf, int outbuflen, mod_cas2_conf *c)
{
  int err, b, ret, total;
  int s = 0;
  struct sockaddr_in sa;
  struct hostent h, *hp2;
  SSL_CTX *ctx = NULL;
  SSL *ssl = NULL;
  X509 *s_cert = NULL;
  char buf[4096];
  SSL_METHOD *method;
  char *full_request, *str, *tmp;

  LOG("in CAS_validate_SSL()\n") ;

  SSLeay_add_ssl_algorithms();
  method = SSLv23_client_method();
  SSL_load_error_strings();
  ctx = SSL_CTX_new(method);
  if (!ctx) FAIL;
  if ((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) FAIL;

  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  hp2 = gethostbyname(c->CASHost) ;
  memcpy(&h, hp2, sizeof(h));
  gethostbyname_r(c->CASHost, &h, buf, sizeof(buf), &b); 
  memcpy(&(sa.sin_addr.s_addr), h.h_addr_list[0], sizeof(long));
  sa.sin_port = htons(atoi(c->CASValidatePort)) ;
  if (connect(s, (struct sockaddr*) &sa, sizeof(sa)) == -1)  FAIL;
  if (!(ssl = SSL_new(ctx))) FAIL;
  SSL_set_fd(ssl, s);
  if (!SSL_connect(ssl)) FAIL;
  if ((s_cert = SSL_get_peer_certificate(ssl)) == NULL) FAIL_BAD_CERT;
  if (!valid_cert(s_cert, c->CASHost, c->CASTrustedCA, c->CASTrustedCADir)) FAIL_BAD_CERT;
  X509_free(s_cert);

  full_request = malloc(strlen(c->CASMethod) + strlen(" ")
    + strlen(c->CASValidate) + strlen("?ticket=") + strlen(ticket) + 
    + strlen("&service=") + strlen(service) + strlen(" ") 
    + strlen(c->CASPort) + strlen("\n\n") + 1); 
  sprintf(full_request, "%s %s?ticket=%s&service=%s %s\n\n",
	  c->CASMethod, c->CASValidate, ticket, service, c->CASProtocol);
  LOG(full_request) ;
  if (!SSL_write(ssl, full_request, strlen(full_request))) FAIL;

  total = 0;
  do {
    b = SSL_read(ssl, buf + total, (sizeof(buf) - 1) - total);
    total += b;
  } while (b > 0);
  buf[total] = '\0';

  // unexpected read error or response too large
  if (b != 0 || total >= sizeof(buf) - 1) FAIL;

  {
    int r = check_cas_reply(buf, outbuf, c->CASMatchValidUser) ;
    END(r) ;
    // SUCCEED;
  }

   /* cleanup and return */

end:
  if (ssl)
    SSL_shutdown(ssl);
  if (s > 0)
    close(s);
  if (ssl)
    SSL_free(ssl);
  if (ctx)
    SSL_CTX_free(ctx);
  return ret;
}

static X509 *get_cert_from_file(char *filename) {
    FILE *f = fopen(filename, "r");
    if (f) {
	X509 *c = PEM_read_X509(f, NULL, NULL, NULL);
	fclose(f);
	return c;
    } else {
	return NULL;
    }
}

#include "log.c"
